/*
 * Copyright (c) 1995, 1996, 1997, 1998, 1999 by Sun Microsystems, Inc.
 * All rights reserved.
 */

#pragma ident	"@(#)shifterCB.cc	5.15	99/07/29 SMI; TSOL 2.x"


#include <stdio.h>

#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/Shell.h>
#include <Xm/List.h>
#include <Xm/Text.h>

#include <tsol/user.h>
#include <tsol/prof.h>
#include <tsol/auth.h>

#include "shifterInclude.hh"

#include<nl_types.h>
extern nl_catd catd;

static int get_roles(item_t **, int *, int);
static int get_profiles(item_t **, int *, int);
static int get_audit_classes(item_t **, int *);
static void shift_selected(item_t *, Widget, Widget);
static void ok_cb(Widget, XtPointer, XtPointer);
static item_t *new_item();
static void free_item(item_t *);
static int itemcmp(const void *, const void *);
static char *profent_comment(profent_t *);
static char *role_comment(userent_t *);

shifter_c::shifter_c()
{
	nitems = 0;
	items = NULL;
}

void
shifter_c::reset()
{
	item_t *p_item;

	XmListDeleteAllItems(left_list);
	XmListDeleteAllItems(right_list);
	for (p_item = items; p_item != NULL; p_item = p_item->p_next)
		XmListAddItemUnselected(left_list, p_item->key_xms, 0);
}

void
shifter_c::role_mode()
{
	XmString tmp_xms;

	mode = MODE_ROLE;

	/* Set Left List Header */
	tmp_xms = XmStringCreateLocalized(catgets(catd,6,3007,"Prohibited"));
	XtVaSetValues(left_lbl, XmNlabelString, tmp_xms, NULL);
	XmStringFree(tmp_xms);

	/* Set Left Right Header */
	tmp_xms = XmStringCreateLocalized(catgets(catd,6,3008,"Assumable"));
	XtVaSetValues(right_lbl, XmNlabelString, tmp_xms, NULL);
	XmStringFree(tmp_xms);

	/* Set Items */
	get_roles(&items, &nitems, TSOL_DB_SRC_SWITCH);
	sort(); 
	reset();
}

void
shifter_c::profile_mode()
{
	XmString tmp_xms = NULL;

	mode = MODE_PROFILE;

	/* Set Left List Header */
	tmp_xms = XmStringCreateLocalized(catgets(catd,6,3000,"Available"));
	XtVaSetValues(left_lbl, XmNlabelString, tmp_xms, NULL);
	XmStringFree(tmp_xms);

	/* Set Left Right Header */
	tmp_xms = XmStringCreateLocalized(catgets(catd,6,3001,"Selected"));
	XtVaSetValues(right_lbl, XmNlabelString, tmp_xms, NULL);
	XmStringFree(tmp_xms);

	/* Set Items */
	nitems = 0;
	free_item(items);
	items = NULL;
	get_profiles(&items, &nitems, TSOL_DB_SRC_SWITCH);
	sort();
	reset();
}

void
shifter_c::select_items(char **itemlist)
{
	char **p_list = NULL;
	item_t *p_item;

	for (p_list = itemlist; *p_list != NULL; p_list++) {
		for (p_item = items; p_item != NULL; p_item = p_item->p_next) {
			if (strcmp(*p_list, p_item->key_txt) == 0) {
				XmListDeleteItem(left_list, p_item->key_xms);
				XmListAddItemUnselected(right_list,
					p_item->key_xms, 0);
				break;
			}
		}
	}
}

char **
shifter_c::get_selected_item_text()
{
	Arg al[10];
	int ac = 0;
	XmString *selected = NULL;
	int count = 0;
	char **names;
	int i = 0;
	int j = 0;

	ac = 0;
	XtSetArg(al[ac], XmNitems, &selected); ac++;
	XtSetArg(al[ac], XmNitemCount, &count); ac++;
	XtGetValues(right_list, al, ac);

	if (count == 0)
		return (NULL);

	if ((names = (char **)malloc((count + 1) * sizeof(char *))) == NULL) {
		return (NULL);
	}
	for (i=0; i<count; i++) {
		XmStringGetLtoR(selected[i], XmFONTLIST_DEFAULT_TAG, &names[i]);
	}
	names[i] = NULL;
	return (names);
}

/*
 * local_auth():
 *   Determine if an authorization is present in any of the selected profiles.
 *   Return Codes:
 *     0: No
 *     1: Yes
 */
int
shifter_c::local_auth(auth_t authno)
{
	item_t *p_item;
	auth_set_t *p_as;

	for (p_item = items; p_item != NULL; p_item = p_item->p_next) {
		if (XmListItemExists(right_list, p_item->key_xms))
		for (p_as=p_item->p_auth; p_as!=NULL; p_as=p_as->p_next)
			if (p_as->auth == authno || p_as->auth == -1)
				return (1);
	}
	return (0);
}

static int
get_roles(item_t **list, int *number, int src)
{
	item_t *p_new = NULL;
	userent_t *p_ue = NULL;
	item_t *p_item = NULL;
	item_t *p_list = NULL;

	*number = 0;
	setuserent(1, src);
	while ((p_ue = getuserent(src)) != NULL) {
		if (strcmp(p_ue->usertype, TSOL_USER_TYPE_NORMAL_KW)) {
			for (p_list = *list; p_list != NULL;
			    p_list = p_list->p_next) { 
				if (!strcmp(p_list->key_txt, p_ue->name))
					break;
			}
			if (!p_list) {
				if ((p_new = new_item()) != NULL) {
					p_new->key_xms = XmStringCreateSimple(p_ue->name);
					p_new->key_txt = strdup(p_ue->name);
					p_new->cmt_txt = role_comment(p_ue);
					p_new->last = 0;
					p_new->p_next = NULL;
					if (*list == NULL) {
						*list = p_new;
						p_item = p_new;
					} else {
						p_item->p_next = p_new;
						p_item = p_new;
					}
				}
				(*number)++;
			}
		}
		free_userent(p_ue);
	}
	return (0);
}

static int
get_profiles(item_t **list, int *number, int src)
{
	item_t *p_new = NULL;
	profent_t *p_pe;
	item_t *p_item;
	item_t *p_list;

	setprofent(1, src);
	while ((p_pe = getprofent(src)) != NULL) {
		for (p_list = *list; p_list != NULL; p_list = p_list->p_next) {
			if (!strcmp(p_list->key_txt, p_pe->name))
				break;
		}
		if (!p_list) {
			if ((p_new = new_item()) != NULL) {
				p_new->key_xms =
				    XmStringCreateLocalized(p_pe->name);
				p_new->key_txt = strdup(p_pe->name);
				p_new->cmt_txt = profent_comment(p_pe);
				p_new->p_auth = str_to_auth_set(p_pe->auths, ",");
				p_new->last = 0;
				if (!strncmp("*",
				    (const char *)p_pe->actions->actname, 3))
					p_new->last++;
				if (!strncmp("*",
				    (const char *)p_pe->cmds->dir, 3))
					p_new->last++;
				p_new->p_next = NULL;
				if (*list == NULL) {
					*list = p_new;
					p_item = p_new;
				} else {
					p_item->p_next = p_new;
					p_item = p_new;
				}
			}
			(*number)++;
		}
		free_profent(p_pe);
	}
	endprofent(src);
	return (0);
}

static int
get_audit_classes(item_t **items, int *number)
{
	items = items;
	*number = *number;

	return (0);
}

void
shifter_c::right(Widget w, XtPointer cd)
{
	XmStringTable xmstrings;
	int count = 0;
	int where;
	int i;

	w = w;
	cd = cd;

	XtVaGetValues(left_list,
	    XmNselectedItems, &xmstrings,
	    XmNselectedItemCount, &count,
	    NULL);

	for (i = 0; i < count ; i++) {
		if (mode == MODE_PROFILE)
			where = profile_position(xmstrings[i]);
		else
			where = 0;
		XmListAddItemUnselected(right_list, xmstrings[i], where);
	}
	XmListDeleteItems(left_list, xmstrings, count);
	XtVaSetValues(right_arrow, XmNsensitive, False, NULL);
	XtVaSetValues(left_arrow, XmNsensitive, False, NULL);
	XtVaSetValues(up_arrow, XmNsensitive, False, NULL);
	XtVaSetValues(down_arrow, XmNsensitive, False, NULL);
}

void
shifter_c::left(Widget w, XtPointer cd)
{
	XmStringTable xmstrings;
	int count = 0;
	int where;
	int i;

	w = w;
	cd = cd;

	XtVaGetValues(right_list,
	    XmNselectedItems, &xmstrings,
	    XmNselectedItemCount, &count,
	    NULL);

	for (i = 0; i < count ; i++) {
		where = sort_position(xmstrings[i]);
		XmListAddItemUnselected(left_list, xmstrings[i], where);
	}
	XmListDeleteItems(right_list, xmstrings, count);
	XtVaSetValues(right_arrow, XmNsensitive, False, NULL);
	XtVaSetValues(left_arrow, XmNsensitive, False, NULL);
	XtVaSetValues(up_arrow, XmNsensitive, False, NULL);
	XtVaSetValues(down_arrow, XmNsensitive, False, NULL);
}

void
shifter_c::up(Widget w, XtPointer cd)
{
	XmString *sel_items;
	XmString selected_xms;
	XmString *all_items;
	int all_count;
	XmString before_xms;
	item_t *p_before_item;
	item_t *p_selected_item;
	int sel_count;
	int pos;

	w = w;
	cd = cd;

	XtVaGetValues(right_list,
	    XmNselectedItems, &sel_items,
	    XmNselectedItemCount, &sel_count,
	    XmNitems, &all_items,
	    XmNitemCount, &all_count,
	    NULL);

	if (!sel_count)
		return;

	selected_xms = XmStringCopy(sel_items[0]);
	pos = XmListItemPos(right_list, selected_xms);
	if (pos == 1)
		return;
	before_xms = XmStringCopy(all_items[pos - 2]);
	p_selected_item = find(selected_xms);
	p_before_item = find(before_xms);
	if (p_selected_item->last != p_before_item->last)
		return;
	XmListDeselectAllItems(right_list);
	XmListDeleteItem(right_list, selected_xms);
	XmListAddItem(right_list, selected_xms, pos - 1);
	XmListSelectItem(right_list, selected_xms, False);
	XmStringFree(selected_xms);
	XmStringFree(before_xms);
}

void
shifter_c::down(Widget w, XtPointer cd)
{
	XmString *sel_items;
	XmString selected_xms;
	XmString *all_items;
	int all_count;
	XmString before_xms;
	item_t *p_before_item;
	item_t *p_selected_item;
	int sel_count;
	int pos;

	w = w;
	cd = cd;

	XtVaGetValues(right_list,
	    XmNselectedItems, &sel_items,
	    XmNselectedItemCount, &sel_count,
	    XmNitems, &all_items,
	    XmNitemCount, &all_count,
	    NULL);

	if (!sel_count)
		return;

	selected_xms = XmStringCopy(sel_items[0]);
	pos = XmListItemPos(right_list, selected_xms);
	if (pos == all_count)
		return;
	before_xms = XmStringCopy(all_items[pos]);
	p_selected_item = find(selected_xms);
	p_before_item = find(before_xms);
	if (p_selected_item->last != p_before_item->last)
		return;
	XmListDeleteItem(right_list, selected_xms);
	XmListAddItem(right_list, selected_xms, pos + 1);
	XmListDeselectAllItems(right_list);
	XmListSelectItem(right_list, selected_xms, False);
	XmStringFree(selected_xms);
	XmStringFree(before_xms);
}

#ifdef DEBUG
void
shifter_c::ok(Widget w, XtPointer cd)
{
	w = w;
	cd = cd;

	char **p_selected = NULL;

	for (p_selected = get_selected_item_text(); p_selected != NULL &&
	    *p_selected != NULL; p_selected++) {
		printf("%s\n", *p_selected);
	}
}
#endif /* DEBUG */

void
shifter_c::audit_mode()
{
	XmString tmp_xms = NULL;

	/* Set Left List Header */
	tmp_xms = XmStringCreateSimple("Classes Ignored");
	XtVaSetValues(left_lbl, XmNlabelString, tmp_xms, NULL);
	XmStringFree(tmp_xms);

	/* Set Left Right Header */
	tmp_xms = XmStringCreateSimple("Classes Audited");
	XtVaSetValues(right_lbl, XmNlabelString, tmp_xms, NULL);
	XmStringFree(tmp_xms);

	/* Set Items */
	get_audit_classes(&items, &nitems);
	sort();
	reset();
}

void
shifter_c::set_src(int newsrc)
{
	switch (newsrc) {
		case TSOL_DB_SRC_NISPLUS:
		case TSOL_DB_SRC_FILES:
		case TSOL_DB_SRC_SWITCH:
			src = newsrc;
			break;
		default:
			src = TSOL_DB_SRC_SWITCH;
	}
}

item_t *
new_item()
{
	item_t *p_new = (item_t *)malloc(sizeof (item_t));
	if (p_new != NULL)
		memset(p_new, NULL, sizeof(item_t));
	return (p_new);
}

void
free_item(item_t *p_item)
{
	item_t *p_next;

	if (p_item == NULL)
		return;
	p_next = p_item->p_next;
	XmStringFree(p_item->key_xms);
	XtFree(p_item->key_txt);
	XtFree(p_item->cmt_txt);
	free_item(p_next);
	free(p_item);
}

/*
 * item_position():
 *   Return the position of xms in the linked list of items.
 *   Assume the first position is numbered 1.
 *   Return 0 if not found.
 */
int
shifter_c::item_position(XmString xms)
{
	int position = 0;
	item_t *p_item;

	for (p_item = items; p_item != NULL; p_item = p_item->p_next) {
		position++;
		if (XmStringCompare(p_item->key_xms, xms))
			break;
	}
	if (p_item)
		return (position);
	return(0);
}

int
shifter_c::profile_position(XmString xms)
{
	XmString *selected = NULL;
	item_t *p_item = NULL;
	item_t *p_insert = NULL;
	int count = 0;
	int i;

	XtVaGetValues(right_list,
	    XmNitems, &selected,
	    XmNitemCount, &count,
	    NULL);

	if (count == 0)
		return(1);

	p_insert = find(xms);

	/*
	 * Find the first item in the right list that has a matching last
	 */
	for (i=0; i<count; i++) {
		p_item = find(selected[i]);
		if (p_insert->last == p_item->last)
			break;
	}

	/*
	 * If i == count, then there was no item with a matching last.
	 */
	if (i == count) {
		if (p_insert->last == 0)
			return (1);
		if (p_insert->last == 2)
			return (0);
		for (i=0; i<count; i++) {
			p_item = find(selected[i]);
			if (p_item->last == 2)
				break;
		}
		return (i+1);
	}

	/*
	 * Since i != count, count is now pointing at the first item with
	 * a matching last value.  Now we need to scan down to the first
	 * item with a non-matching last, and insert before that.
	 */
	for (i=i+1; i<count; i++) {
		p_item = find(selected[i]);
		if (p_insert->last != p_item->last)
			break;
	}
	return (i+1);
}

int
shifter_c::sort_position(XmString xms)
{
	item_t *p_item;
	item_t *p_insert;
	XmString *selected;
	int count, i;

	XtVaGetValues(left_list,
	    XmNitems, &selected,
	    XmNitemCount, &count,
	    NULL);

	p_insert = find(xms);
	for (i = 0; i < count; i++) {
		p_item = find(selected[i]);
		if (strcmp(p_insert->key_txt, p_item->key_txt) < 0) {
			return (i + 1);
		}
	}
	return (0);
}

item_t *
shifter_c::find(XmString xms)
{
	item_t *p_item;

	for (p_item = items; p_item != NULL; p_item = p_item->p_next)
		if (XmStringCompare(xms, p_item->key_xms))
			return (p_item);
	return (NULL);
}

void
shifter_c::sort()
{
	int i;
	item_t **item_array;
	item_t *p_item;

	item_array = (item_t **)malloc(nitems * sizeof (item_t *));
	if (item_array == NULL)
		return;
	for (i = 0, p_item = items; i < nitems; i++, p_item = p_item->p_next)
		item_array[i] = p_item;
	qsort(item_array, (size_t)nitems, sizeof (item_t *), itemcmp);
	items = item_array[0];
	for (i = 1, p_item = items; i < nitems;
	    i++, p_item = p_item->p_next)
		p_item->p_next = item_array[i];
	p_item->p_next = NULL;
	free(item_array);
}

static int
itemcmp(const void *i1, const void *i2)
{
	item_t *it1 = *(item_t **)i1;
	item_t *it2 = *(item_t **)i2;

	return (strcmp(it1->key_txt, it2->key_txt));
}

void
shifter_c::comment(Widget w, XtPointer cd)
{
	XmString *selected;
	int count;
	item_t *p_item;

	cd = cd;

	XtVaGetValues(w,
	    XmNselectedItems, &selected,
	    XmNselectedItemCount, &count,
	    NULL);

	if (!count)
		return;

	p_item = find(selected[0]);
	XmTextSetString(text, p_item->cmt_txt);
}

void
shifter_c::deselect_opposite(Widget w, XtPointer cd)
{
	cd = cd;

	if (w == right_list) {
		XmListDeselectAllItems(left_list);
		XtVaSetValues(right_arrow, XmNsensitive, False, NULL);
		XtVaSetValues(left_arrow, XmNsensitive, True, NULL);
		XtVaSetValues(up_arrow, XmNsensitive, True, NULL);
		XtVaSetValues(down_arrow, XmNsensitive, True, NULL);
	} else {
		XmListDeselectAllItems(right_list);
		XtVaSetValues(right_arrow, XmNsensitive, True, NULL);
		XtVaSetValues(left_arrow, XmNsensitive, False, NULL);
		XtVaSetValues(up_arrow, XmNsensitive, False, NULL);
		XtVaSetValues(down_arrow, XmNsensitive, False, NULL);
	}
}

/*
 * profent_comment():
 *   Format comment from profent structure.
 *   Comment is of the form:
 *
 *   Prof: <profile-name>
 *   Desc: <description>
 *   Auth: <authorizations>
 *   Acts: <action-names>
 *   Cmds: <command-names>
 */
static char *
profent_comment(profent_t *p_pe)
{
	int len = 1;
	profact_t *p_act;
	profcmd_t *p_cmd;
	auth_set_t *p_al;
	char *authstr;
	char *cmt;

	char *prof_txt = catgets(catd,6,3002,"Prof: ");
	char *desc_txt = catgets(catd,6,3003,"Desc: ");
	char *auth_txt = catgets(catd,6,3004,"Auth: ");
	char *acts_txt = catgets(catd,6,3005,"Acts: ");
	char *cmds_txt = catgets(catd,6,3006,"Cmds: ");

	len += strlen(prof_txt) + 1;
	len += strlen(desc_txt) + 1;
	len += strlen(auth_txt) + 1;
	len += strlen(acts_txt) + 1;
	len += strlen(cmds_txt) + 1;

	len += strlen(p_pe->name) + 1;
	len += strlen(p_pe->desc) + 1;
	p_al = str_to_auth_set(p_pe->auths, ",");
	authstr = auth_set_to_str(p_al, ',');
	len += strlen(authstr) + 1;
	for (p_act = p_pe->actions; p_act != NULL; p_act = p_act->p_next)
		len += strlen(p_act->actname) + 2;
	for (p_cmd = p_pe->cmds; p_cmd != NULL; p_cmd = p_cmd->p_next) {
		if (p_cmd->file && p_cmd->file[0] != NULL)
			len += strlen(p_cmd->file) + 2;
		else if (p_cmd->dir)
			len += strlen(p_cmd->dir) + 2;
	}

	if ((cmt = (char *)malloc(sizeof (char) * len)) == NULL)
		return (strdup(""));
	sprintf(cmt, "%s%s\n%s%s\n%s%s\n%s",
	    prof_txt, p_pe->name,
	    desc_txt, p_pe->desc,
	    auth_txt, authstr, 
	    acts_txt);
	for (p_act = p_pe->actions; p_act != NULL; p_act = p_act->p_next) {
		strcat(cmt, p_act->actname);
		if (p_act->p_next)
			strcat(cmt, ", ");
	}
	strcat(cmt, "\n");
	strcat(cmt, cmds_txt);
	for (p_cmd = p_pe->cmds; p_cmd != NULL; p_cmd = p_cmd->p_next) {
		if (p_cmd->file && p_cmd->file[0] != NULL)
			strcat(cmt, p_cmd->file);
		else if (p_cmd->dir)
			strcat(cmt, p_cmd->dir);
		if (p_cmd->p_next)
			strcat(cmt, ", ");
	}
	strcat(cmt, "\n");
	return (cmt);
}

/*
 * role_comment():
 *   Format comment from profent structure.
 *   Comment is of the form:
 *
 *   Role: <username>
 *   Prof: <profiles>
 */
static char *
role_comment(userent_t *p_ue)
{
	char *cmt;
	int len;

	char role_txt[50];
	char profile_txt[50];

	strcpy(role_txt,catgets(catd,6,3009, "Role: "));
	strcpy(profile_txt,catgets(catd,6,3010, "Profile: "));

	len = strlen(role_txt) + strlen(p_ue->name) +
	    strlen(profile_txt) + strlen(p_ue->profiles) + 2;
	cmt = (char *)malloc(len * sizeof (char));
	if (cmt) {
		sprintf(cmt, "%s%s\n%s%s", role_txt, p_ue->name,
		    profile_txt, p_ue->profiles);
	} else {
		cmt = strdup("");
	}

	return (cmt);
}

void
shifter_c::add_item(char *s)
{
	item_t *p_new;

	if ((p_new = new_item()) == NULL)
		return;

	p_new->key_xms = XmStringCreateSimple(s);
	p_new->key_txt = strdup(s);
	p_new->cmt_txt = strdup(catgets(catd,6,3011,"New Role"));
	p_new->last = 0;
	p_new->p_next = items;
	items = p_new;
	nitems++;
	sort();
}
